#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : detect.py
# Contents: detect synchronous spiking activity and reduce results
# Author  : Christian Borgelt
# History : 2013.02.13 file created from detect/detect.py
#           2013.02.23 added jitter to injected coincidences
#           2013.02.25 extended to larger grid 2,...,12 x 2,...,12
#           2013.02.27 combined with script for false negatives
#           2013.04.29 jitter added to output file names
#           2013.05.07 added false negative rates for raw patterns
#           2013.05.08 adapted some file names, removed filtered result
#           2013.12.06 adapted to filtering with detection border
#           2014.05.27 modules placed in sibling directories
#           2015.08.12 adapted to modified pattern set reduction
#           2015.09.09 completed adaptation to cover similarity
#-----------------------------------------------------------------------
from sys     import argv, path
from os      import makedirs
from os.path import join
from time    import strftime
from random  import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par  import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts    import genpsts
from ccnovl  import cocojim

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------

def redfn00 (pats):
    return pats

def redfn01 (pats):
    return patred([(p,c, len(p)   *c)               for p,c in pats])

def redfn02 (pats):
    return patred([(p,c,(len(p)-1)*c)               for p,c in pats])

def redfn03 (pats):
    return patred([(p,c,(len(p)-1)*(c+1.0))         for p,c in pats])

def redfn04 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.10*len(p))) for p,c in pats])

def redfn05 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.12*len(p))) for p,c in pats])

def redfn06 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.14*len(p))) for p,c in pats])

def redfn07 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.15*len(p))) for p,c in pats])

def redfn08 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.16*len(p))) for p,c in pats])

redfns = [redfn00, redfn01, redfn02, redfn03, redfn04,
          redfn05, redfn06, redfn07, redfn08]

#-----------------------------------------------------------------------

def patred (pats):
    pats.sort(key=lambda x: x[2], reverse=True)
    pall = pats                 # note the set of all patterns
    tmp  = []                   # initialize intermediary pattern list
    while pats:                 # while there are more patterns
        tmp.append(pats[0])     # transfer best pattern to output
        p,c,v = pats[0]         # remove all subset patterns
        pats = [q for q in pats[1:] if not q[0] <= p]
    for p,c,v in tmp:           # traverse the intermediary patterns
        for q,x,z in pall:      # traverse all patterns
            if z < v: pats.append((p,c)); break
            if q < p: break     # remove all supersets that are worse
    return pats                 # return the reduced pattern list

#-----------------------------------------------------------------------

def detect (ncnt=100, period=3, rate=20,         # neuron    parameters
            size=5, coin=5, jitter=0.002,        # assembly  parameters
            width=0.003, simcov='j',             # detection parameters
            target='c', rfnids=[0],              # detection parameters
            runs=1000, verbose=100, srng=1):     # other     parameters
    if jitter < 0: jitter = width
    srng += 1023*size           # init. seed and get data file name
    base = 'n%d_p%g_r%g_w%g_%c_%c' \
         % (ncnt,period,rate,width*10000,simcov,target)
    border = [0 for i in range(ncnt+1)]
    with open(join('data','pc','bd_'+base+'.txt'), 'r') as inp:
        for line in inp:        # collect the support values
            z,m = line.split(); border[int(z)] = float(m)
    for z in range(ncnt-1,-1,-1): # ensure a monotone decision border
        if border[z+1] > border[z]: border[z] = border[z+1]
    base = 'n%d_p%g_r%g_j%g_w%g_%c_%c' \
         % (ncnt,period,rate,jitter*10000,width*10000,simcov,target)
    ext  = '_%d_%d' % (size,coin)
    name = base +ext            # form the file name (parts)
    asm  = set(range(1,size+1)) # create the assembly to inject
    upcs = dict()               # and the pattern count dictionaries
    fpcs = dict()               # (unfiltered, filtered, reduced)
    rpcs = dict((r,dict()) for r in rfnids)
    draw = xraw = 0                     # runs with raw detections
    dect = dict((r,0) for r in rfnids)  # runs with detections
    xdct = dict((r,0) for r in rfnids)  # runs with exact detections
    rexs = dict((r,0) for r in rfnids)  # runs with excess  neurons
    rmis = dict((r,0) for r in rfnids)  # runs with missing neurons
    rops = dict((r,0) for r in rfnids)  # runs with other     patterns
    sups = dict((r,0) for r in rfnids)  # number of superset  patterns
    subs = dict((r,0) for r in rfnids)  # number of subset    patterns
    ovls = dict((r,0) for r in rfnids)  # number of overlap   patterns
    urls = dict((r,0) for r in rfnids)  # number of unrelated patterns
    for i in range(runs):       # repeat the experiment multiple times
        if verbose > 0 and i % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S detect ')+name+' '+str(i))
        seed(srng)              # seed random number generator
        k = int(ncnt/4); n = ncnt -3*k
        psts = genpsts(k, period, int(rate*0.4), jitter=jitter,
                       asms=[[asm,-coin]]) \
             + genpsts(k, period, int(rate*0.8)) \
             + genpsts(k, period, int(rate*1.2)) \
             + genpsts(n, period, int(rate*1.6))
        pats = cocojim(list(zip(range(1,ncnt+1), psts)),
                       target=target, width=width, supp=1.0,
                       simcov=simcov, sim=0, zmin=2, zmax=-1,
                       mode='e', report='aw')
        fltd = []; d = e = 0    # generate and mine spike trains
        for p,c,w in pats:      # and traverse the found patterns
            z = len(p)          # get the size of the pattern
            p = frozenset(p)    # and the set of items/neurons
            if p >= asm:        # check for raw detections
                if z > size: e = 1     # superset detection
                else:        d = 1     # exact    detection
            s = (z,int(c*1000)) # compute histogram bin
            if s in upcs: upcs[s] += 1 # count unfiltered patterns
            else:         upcs[s]  = 1 # by their signatures
            if c <= border[z]: continue
            if s in fpcs: fpcs[s] += 1 # count filtered patterns
            else:         fpcs[s]  = 1 # by their signatures
            fltd.append((p,c if simcov == 'x' else w))
        draw += d | e           # collect non-excluded patterns and
        xraw += d               # count runs with raw detections
        for r in rfnids:        # traverse the reduction function ids
            pats  = redfns[r](fltd)
            pcs   = rpcs[r]     # reduce the filtered patterns
            d = e = x = y = 0   # initialize the detection flags
            for p,c in pats:    # traverse the reduced patterns
                z = len(p)      # get pattern size and
                s = (z,int(c*1000))      # compute histogram bin
                if s in pcs: pcs[s] += 1 # count reduced patterns
                else:        pcs[s]  = 1 # by their signatures
                if   p >= asm:  # set detection flags, count patterns
                    if z > size:       sups[r] += 1; e = 1
                    else:                            d = 1
                elif p <  asm:         subs[r] += 1; x = 1
                elif len(p & asm) > 1: ovls[r] += 1; x = 1
                else:                  urls[r] += 1; x = 1
            dect[r] += d | e    # count runs with detections,
            xdct[r] += d        # runs with exact detections,
            rexs[r] += e        # runs with excess neurons,
            rmis[r] += x        # runs with no detect patterns
            rops[r] += x | e    # runs with all other patterns
        srng += coin            # switch to the next random seed
    if verbose > 0:             # print termination log message
        print(strftime('%Y.%m.%d %H:%M:%S detect ')
                      +name+' '+str(runs)+' done.')
    for subdir in [join('raw','x'),'pu','pf','pr']:
        try: makedirs(join('data',subdir))
        except OSError: pass    # ensure that subdirectories exist
    n = 1.0/runs if runs > 0 else 1.0
    with open(join('data','raw','x','fn_'+name+'.txt'), 'w') as out:
        out.write('%g '  % (1.0 -draw*n))
        out.write('%g\n' % (1.0 -xraw*n))
    with open(join('data','pu','pu_'+name+'.txt'), 'w') as out:
        for x in sorted((s[0], s[1], upcs[s]*n) for s in upcs):
            out.write('%g %g %g\n' % x)
    with open(join('data','pf','pf_'+name+'.txt'), 'w') as out:
        for x in sorted((s[0], s[1], fpcs[s]*n) for s in fpcs):
            out.write('%g %g %g\n' % x)
    for r in rfnids:            # traverse the reduction function ids
        name = base+str(r)+ext  # build the file name and
        pcs  = rpcs[r]          # write the pattern counters
        try: makedirs(join('data','pr',str(r)))
        except OSError: pass    # ensure that directory exists
        outfn = join('data','pr',str(r),'pr_'+name+'.txt')
        with open(outfn, 'w') as out:
            for x in sorted((s[0], s[1], pcs[s]*n) for s in pcs):
                out.write('%g %g %g\n' % x)
        try: makedirs(join('data','raw',str(r)))
        except OSError: pass    # ensure that directory exists
        outfn = join('data','raw',str(r),'fn_'+name+'.txt')
        with open(outfn, 'w') as out:
            out.write('%g '  % (1.0 -dect[r]*n))
            out.write('%g '  % (1.0 -xdct[r]*n))
            out.write('%g '  % (rexs[r]*n)) # write the average
            out.write('%g '  % (rmis[r]*n)) # pattern counts
            out.write('%g '  % (rops[r]*n))
            out.write('%g '  % (sups[r]*n))
            out.write('%g '  % (subs[r]*n))
            out.write('%g '  % (ovls[r]*n))
            out.write('%g\n' % (urls[r]*n))
    return (upcs, fpcs, rpcs,   # return the pattern spectrums
            dect, xdct, rexs, rmis, rops, sups, subs, ovls, urls)

#-----------------------------------------------------------------------

def collect (ncnt=100, period=3, rate=20,        # neuron    parameters
             sizes=[5], coins=[5], jitter=0.002, # assembly  parameters
             width=0.003, simcov='j',            # detection parameters
             target='c', rfnids=[4]):            # detection parameters
    if jitter < 0: jitter = width
    base = 'n%d_p%g_r%g_j%g_w%g_%c_%c' \
         % (ncnt,period,rate,jitter*10000,width*10000,simcov,target)
    fneg = dict()               # get the pattern count file name base
    fnxs = dict()               # and initialize run count dictionaries
    for s in [(z,c) for z in sizes for c in coins]:
        inpfn = join('data','raw','x','fn_'+base+('_%d_%d.txt' % s))
        with open(inpfn, 'r') as inp:
            fneg[s],fnxs[s] = inp.read()[:-1].split()
    for subdir in ['fn','fx','rx','rn','ro','sp','sb','ov','ur']:
        try: makedirs(join('data',subdir))
        except OSError: pass    # ensure that subdirectories exist
    with open(join('data','fn','fn_' +base +'.txt'), 'w') as out:
        for x in sorted((k[0], k[1], fneg[k]) for k in fneg):
            out.write('%g %g %s\n' % x)
    with open(join('data','fx','fx_' +base +'.txt'), 'w') as out:
        for x in sorted((k[0], k[1], fnxs[k]) for k in fnxs):
            out.write('%g %g %s\n' % x)
    for r in rfnids:            # traverse the reduction function ids
        name = base +str(r)     # get the pattern count file name
        fneg = dict(); fnxs = dict()
        rexs = dict(); rmis = dict(); rops = dict()
        sups = dict(); subs = dict(); ovls = dict(); urls = dict()
        for s in [(z,c) for z in sizes for c in coins]:
            outfn = join('data','raw',str(r),
                         'fn_' +name +('_%d_%d.txt' % s))
            with open(outfn, 'r') as inp:
                vals = inp.read()[:-1].split()
                fneg[s],fnxs[s]                 = vals[0:2]
                rexs[s],rmis[s],rops[s]         = vals[2:5]
                sups[s],subs[s],ovls[s],urls[s] = vals[5:9]
        with open(join('data','fn','fn_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], fneg[k]) for k in fneg):
                out.write('%g %g %s\n' % x)
        with open(join('data','fx','fx_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], fnxs[k]) for k in fnxs):
                out.write('%g %g %s\n' % x)
        with open(join('data','rx','rx_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], rexs[k]) for k in rexs):
                out.write('%g %g %s\n' % x)
        with open(join('data','rn','rn_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], rmis[k]) for k in rmis):
                out.write('%g %g %s\n' % x)
        with open(join('data','ro','ro_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], rops[k]) for k in rops):
                out.write('%g %g %s\n' % x)
        with open(join('data','sp','sp_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], sups[k]) for k in sups):
                out.write('%g %g %s\n' % x)
        with open(join('data','sb','sb_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], subs[k]) for k in subs):
                out.write('%g %g %s\n' % x)
        with open(join('data','ov','ov_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], ovls[k]) for k in ovls):
                out.write('%g %g %s\n' % x)
        with open(join('data','ur','ur_' +name +'.txt'), 'w') as out:
            for x in sorted((k[0], k[1], urls[k]) for k in urls):
                out.write('%g %g %s\n' % x)

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'target', [  'c' ] ),
                ( {'jitter':0.0020, 'width':0.0030},
                  {'jitter':0.0030, 'width':0.0045} ),
                #( {'jitter':0.0020, 'width':0.0030}, ),
                #( {'jitter':0.0030, 'width':0.0045}, ),
                ( 'simcov', [ 'x', 'j', 'k', 'd', 's' ] ),
                ( 'rfnids', [ [1,2,3,4,5,6,7,8] ] ),
                ( 'runs',   [ 1000 ] ) ] ]

    if len(argv) < 2 or argv[1] != 'collect':
        run(detect, grids, int(argv[1])-1 if len(argv) > 1 else -1)
    else:
        ranges = dict(project(grids, ['size','coin'])[0])
        grids  = project(grids, ['ncnt','period','rate',
                                 'width','jitter','simcov',
                                 'target','rfnids'])
        for i in range(get_args(grids)):
            collect(sizes=ranges['size'], coins=ranges['coin'],
                    **get_args(grids, i))
